from invisibleroads_macros_configuration import set_default
from os import environ
from pyramid.events import NewResponse
from pyramid.httpexceptions import HTTPError
from pyramid.security import NO_PERMISSION_REQUIRED
from pyramid.settings import asbool, aslist

from .constants import POST_ID_LENGTH
from .routines.cache import configure_cache
from .routines.configuration import fill_settings, set_attribute
from .variables import FUNCTION_CACHE, POSTS_REGISTRY
from .views import handle_http_error


class CorsPreflightPredicate(object):

    def __init__(self, value, config):
        self.value = value

    def text(self):
        return 'cors_preflight = %s' % self.value

    phash = text

    def __call__(self, context, request):
        if not self.value:
            return False
        method = request.method
        headers = request.headers
        return (
            'OPTIONS' == method and
            'Origin' in headers and
            'Access-Control-Request-Method' in headers)


def includeme(config):
    configure_settings(config)
    configure_cross_origin_resource_sharing(config)
    configure_caches(config)
    configure_views(config)


def configure_settings(config):
    settings = config.get_settings()
    fill_settings(settings)
    set_attribute(
        settings, POSTS_REGISTRY, 'id.length', POST_ID_LENGTH, int)

    if 'data.folder' in settings:
        config.add_request_method(
            lambda request: settings['data.folder'], 'data_folder', reify=True)

    for line in settings.get(
            'application.environment', '').strip().splitlines():
        k, v = line.split('=', 1)
        k = k.strip()
        v = v.strip()
        environ[k] = v


def configure_cross_origin_resource_sharing(config):
    settings = config.get_settings()
    allowed_origins = aslist(settings.get(
        'cors.allowed_origins', '').casefold())
    if not allowed_origins:
        return
    allowed_methods = aslist(settings.get('cors.allowed_methods', '').upper())
    allowed_headers = aslist(settings.get('cors.allowed_headers', '').title())
    allow_credentials = asbool(settings.get('cors.allow_credentials'))

    def see_options(request):
        response = request.response
        response_headers = response.headers
        response_headers['Access-Control-Allow-Methods'] = ','.join(
            allowed_methods)
        response_headers['Access-Control-Allow-Headers'] = ','.join(
            allowed_headers)
        return response

    def set_cross_origin_resource_sharing(event):
        request = event.request
        request_headers = request.headers
        response = event.response
        response_headers = response.headers
        if allow_credentials:
            response_headers['Access-Control-Allow-Credentials'] = 'true'
        key = 'Access-Control-Allow-Origin'

        if response_headers.get(key) == '*':
            return
        if '*' in allowed_origins:
            response_headers[key] = '*'
            return
        if len(allowed_origins) == 1:
            response_headers[key] = allowed_origins[0]
            return

        origin = request_headers.get('Origin', '')
        if origin.casefold() not in allowed_origins:
            response.headers = {
                k: v for k, v in response_headers.items()
                if not k.startswith('Access-Control-')}
            return
        response_headers[key] = origin
        response_headers['Vary'] = 'Origin'

    # https://gist.github.com/mmerickel/1afaf64154b335b596e4
    config.add_route_predicate('cors_preflight', CorsPreflightPredicate)
    config.add_route('corsPreflight', '/{all:.*}', cors_preflight=True)
    config.add_view(
        see_options, route_name='corsPreflight',
        permission=NO_PERMISSION_REQUIRED)
    config.add_subscriber(set_cross_origin_resource_sharing, NewResponse)


def configure_caches(config):
    settings = config.get_settings()

    # Configure client_cache
    http_expiration_time_in_seconds = set_default(
        settings, 'client_cache.http.expiration_time_in_seconds', 3600, int)

    config.add_directive(
        'add_cached_static_view',
        lambda config, *arguments, **keywords: config.add_static_view(
            *arguments, cache_max_age=http_expiration_time_in_seconds,
            **keywords))

    config.add_directive(
        'add_cached_view',
        lambda config, *arguments, **keywords: config.add_view(
            *arguments, http_cache=http_expiration_time_in_seconds,
            **keywords))

    # Configure server_cache
    configure_cache(settings, 'server_cache.function.', FUNCTION_CACHE)


def configure_views(config):
    config.add_exception_view(
        handle_http_error, context=HTTPError, renderer='json')



from functools import lru_cache
from invisibleroads_macros_configuration import (
    SECRET_LENGTH, fill_environment_variables, fill_extensions, fill_secrets,
    set_default)
from pyramid.paster import bootstrap, get_appsettings, setup_logging


@lru_cache(maxsize=None)
def load_bootstrapped_settings(configuration_path):
    'Load settings with full configuration'
    setup_logging(configuration_path)
    with bootstrap(configuration_path) as env:
        request = env['request']
    return request.registry.settings


@lru_cache(maxsize=None)
def load_filled_settings(configuration_path):
    setup_logging(configuration_path)
    settings = get_appsettings(configuration_path)
    fill_settings(settings)
    return settings


def fill_settings(settings):
    fill_environment_variables(settings)
    fill_secrets(settings, secret_length=set_default(
        settings, 'secret.length', SECRET_LENGTH, int))
    fill_extensions(settings)


def set_attribute(settings, class_registry, attribute_name, default, parse):
    for class_name, Class in class_registry.items():
        if class_name in (
            'EnumeratedBase',
            'RandomBase',
        ) or class_name.startswith('_'):
            continue
        prefix = getattr(Class, '__tablename__', class_name.casefold()) + '.'
        key = prefix + attribute_name
        value = set_default(settings, key, default, parse)
        setattr(Class, attribute_name.replace('.', '_'), value)



from invisibleroads_macros_disk import remove_folder
from pyramid import testing
from pytest import fixture


@fixture
def posts_request(posts_config, data_folder):
    request = testing.DummyRequest(
        data_folder=data_folder)
    request.json_body = {}
    yield request


@fixture
def posts_config(posts_settings):
    config = testing.setUp(settings=posts_settings)
    config.include('invisibleroads_posts')
    yield config
    testing.tearDown()


@fixture
def posts_settings():
    return {
        'data.folder': data_folder,
        'secret.length': 32,
        'client_cache.http.expiration_time_in_seconds': 60,
    }


@fixture
def data_folder(tmpdir):
    data_folder = str(tmpdir)
    yield data_folder
    remove_folder(data_folder)



from invisibleroads_macros_descriptor import classproperty
from invisibleroads_macros_disk import (
    check_absolute_path, make_enumerated_folder, make_random_folder)
from invisibleroads_macros_text import normalize_key
from os.path import basename, exists, join
from pyramid.httpexceptions import HTTPNotFound

from .variables import POSTS_REGISTRY
from .views import get_value


class RegisteredClass(type):

    def __init__(Class, name, *args):
        super().__init__(name, *args)
        POSTS_REGISTRY[name] = Class


class EnumeratedBase(object):

    def __init__(self, **kw):
        self.__dict__.update(kw)


class RandomBase(EnumeratedBase, metaclass=RegisteredClass):
    pass


class FolderMixin(object):

    def get_folder(self, data_folder):
        base_folder = self.get_base_folder(data_folder)
        return check_absolute_path(str(self.id), base_folder)

    @classmethod
    def get_from(Class, request, record_id=None):
        key = Class.singular_descriptor + 'Id'
        if record_id is None:
            record_id = get_value(request, key)
        data_folder = request.data_folder
        record = Class(id=record_id)
        record_folder = record.get_folder(data_folder)
        if not exists(record_folder):
            raise HTTPNotFound({key: 'bad'})
        return record

    @classmethod
    def make_unique_record(Class, data_folder):
        base_folder = Class.get_base_folder(data_folder)
        record = Class()
        try:
            id_length = getattr(Class, 'id_length')
        except AttributeError:
            record_folder = make_enumerated_folder(base_folder)
        else:
            record_folder = make_random_folder(base_folder, id_length)
        record.id = basename(record_folder)
        record.folder = record_folder
        return record

    @classmethod
    def get_base_folder(Class, data_folder):
        return join(data_folder, Class.plural_descriptor)

    @classproperty
    def plural_descriptor(Class):
        return Class.singular_descriptor + 's'

    @classproperty
    def singular_descriptor(Class):
        key = getattr(Class, '__name__', Class.__class__.__name__)
        return normalize_key(key, word_separator='_', separate_camel_case=True)



from pyramid.httpexceptions import HTTPBadRequest


def handle_http_error(request):
    exception = request.exception
    request.response.status_int = exception.status_int
    return exception.args[0]


def expect_integer(
        request, key, parse=None, minimum=None, maximum=None, default=None):
    value = expect_value(
        request, key, int, 'expected integer', default=default)
    if parse:
        try:
            value = parse(value)
        except ValueError:
            raise HTTPBadRequest({key: 'unexpected value'})
    if minimum and value < minimum:
        raise HTTPBadRequest({key: 'expected value >= %s' % minimum})
    if maximum and value > maximum:
        raise HTTPBadRequest({key: 'expected value <= %s' % maximum})
    return value


def expect_value(request, key, parse=None, message=None, default=None):
    value = get_value(request, key, default)
    if parse:
        try:
            value = parse(value)
        except (KeyError, ValueError):
            raise HTTPBadRequest({key: message or 'bad'})
    return value


def get_value(request, key, default=None):
    try:
        return request.matchdict[key]
    except KeyError:
        pass
    try:
        return request.params[key]
    except KeyError:
        pass
    try:
        return request.json_body[key]
    except KeyError:
        pass
    if default is not None:
        return default
    raise HTTPBadRequest({key: 'required'})



from invisibleroads.scripts import ConfigurableScript
from invisibleroads_macros_disk import make_folder, remove_folder

from .routines.configuration import load_filled_settings


class InitializePostsScript(ConfigurableScript):

    priority = 10

    def configure(self, argument_subparser):
        super(InitializePostsScript, self).configure(argument_subparser)
        argument_subparser.add_argument('--restart', action='store_true')

    def run(self, args, argv):
        settings = load_filled_settings(args.configuration_path)
        if args.restart and 'data.folder' in settings:
            remove_folder(settings['data.folder'])
        for key, value in settings.items():
            if key.endswith('.folder'):
                make_folder(value)



def configure_cache(settings, prefix, cache):
    if prefix + 'backend' not in settings:
        settings[prefix + 'backend'] = 'dogpile.cache.memory'
    settings[prefix + 'replace_existing_backend'] = True
    cache.configure_from_config(settings, prefix)
